import math
import numpy as np
import matplotlib.pyplot as plt

class DecisionsEns():
    def __init__(self):
        self.decisionfunctionssig = []
        self.decisionfunctionsnoise = []
        self.sigdfvs = []
        self.sigpmfs = []
        self.noisedfvs = []
        self.noisepmfs = []
        self.pdets = []
        self.pfas = []

    def finalize(self):
        self.globalmin = np.floor(np.min(self.decisionfunctionsnoise))
        self.globalmax = np.ceil(np.max(self.decisionfunctionssig))
        self.decfuncx = np.linspace(self.globalmin, self.globalmax, self.globalmax - self.globalmin)
        for ndx in range(len(self.decisionfunctionssig)):
            sigdfv, sigpmf = self.decisionfunctionpmf(self.decisionfunctionssig[ndx])
            self.sigdfvs.append(sigdfv)
            self.sigpmfs.append(sigpmf)
            noisedfv, noisepmf = self.decisionfunctionpmf(self.decisionfunctionsnoise[ndx])
            self.noisedfvs.append(noisedfv)
            self.noisepmfs.append(noisepmf)
            pdet = np.zeros(len(sigpmf))
            pfa = np.zeros(len(noisepmf))
            lastndx = len(sigpmf) - 1
            for ndx in range(lastndx + 1):
                if ndx == 0:
                    pdet[ndx] = sigpmf[lastndx]
                    pfa[ndx] = noisepmf[lastndx]
                else:
                    pdet[ndx] = pdet[ndx - 1] + sigpmf[lastndx - ndx]
                    pfa[ndx] = pfa[ndx - 1] + noisepmf[lastndx - ndx]
            self.pdets.append(pdet)
            self.pfas.append(pfa)
        pass

    def decisionfunctionpmf(self, vals):
        from pymc3.plots.kdeplot import fast_kde
        vals = np.append(vals, [self.globalmin, self.globalmax])
        pmf, min, max = fast_kde(vals, bw=4.5)
        x = np.linspace(min, max, len(pmf))
        pmf = pmf / sum(pmf)
        return x, pmf

    def plot_roc(self):
        plt.figure()
        for ndx in range(len(self.pdets)):
            plt.plot(self.pfas[ndx], self.pdets[ndx], 'g', alpha=.1)
        plt.xlim(0., 1.)
        plt.ylim(0., 1.)
        plt.xlabel('probability false-alarm')
        plt.ylabel('probability detection')

    def plot_decisionfunctions(self):
        plt.figure()
        for ndx in range(len(self.decisionfunctionssig)):
            plt.plot(self.sigdfvs[ndx], self.sigpmfs[ndx], 'g', alpha=.1)
            plt.plot(self.noisedfvs[ndx], self.noisepmfs[ndx], 'b', alpha=.1)
        sigx, sigpmf = self.decisionfunctionpmf(self.decisionfunctionssig)
        noisex, noisepmf = self.decisionfunctionpmf(self.decisionfunctionsnoise)
        plt.plot(sigx, sigpmf, 'g')
        plt.plot(noisex, noisepmf, 'b')
        plt.xlabel('decision function output-value')
        plt.ylabel('p')

    def addsig(self, decisionfunction):
        self.decisionfunctionssig.append(decisionfunction)

    def addnoise(self, decisionfunction):
        self.decisionfunctionsnoise.append(decisionfunction)

class Decisions():
    """framework for evaluating decisions."""
    def __init__(self, mode, tracker):
        self.mode = mode
        tracker.innov.finalize()
        self.t = tracker.innov.log[:, 0]
        self.yhat = tracker.innov.log[:, 2]
        self.inn = tracker.innov.log[:, 4]
        self.Rhatee = tracker.innov.log[:, 5]
        self.decisionfunction = self.evaldecisionfunction()

    def evaldecisionfunction(self):
        decisionfunction = np.zeros(len(self.inn))
        for ndx in range(len(decisionfunction)):
            delta = (self.yhat[ndx]**2 / 8) - (self.inn[ndx]**2 / (2 * self.Rhatee[ndx]))
            if ndx == 0: decisionfunction[0] = delta
            else: decisionfunction[ndx] = decisionfunction[ndx-1] + delta
        return decisionfunction

def rccircuit(runs):
    from kalman import Classical
    ens = DecisionsEns()
    for runndx in range(runs):
        tracker = Classical(mode='rccircuit', plot=False, signal=300.)
        dec = Decisions(mode='rccircuit', tracker=tracker)
        ens.addsig(decisionfunction=dec.decisionfunction)
        tracker = Classical(mode='rccircuit', plot=False, signal=0.)
        dec = Decisions(mode='rccircuit', tracker=tracker)
        ens.addnoise(decisionfunction=dec.decisionfunction)
    ens.finalize()
    return ens

if __name__ == "__main__":
    ens = rccircuit(runs=100)
    ens.plot_decisionfunctions()
    ens.plot_roc()
    plt.show()
    pass


class Innovs():
    def __init__(self):
        self.log = []

    def update(self, t, xhat, yhat, err, inn):
        self.log.append([t, xhat, yhat, err, inn])

    def update2(self, t, xhat, yhat, err, inn, Ree, Ptil):
        self.log.append([t, xhat, yhat, err, inn, Ree, Ptil])

    def autocorr1(self, x):
        n = x.size
        ac = np.zeros([n, 1])
        for k in range(1, n):  # k = 1 to n-1
            ac[k] = 0
            for t in range(1, n - k):  # t = 1 to n-k
                ac[k] += (self.ets[t - 1] - self.mhate) * (self.ets[t - 1 + k] - self.mhate)  # / (n-k)
        return ac

    def autocorr2(self, x):
        ac = np.correlate(x, x, mode='full')
        return ac[ac.size // 2:]

    def finalize(self):
        self.log = np.asarray(self.log)
        self.tts = self.log[:, 0]
        self.ets = self.log[:, 4]
        self.mhate = np.mean(self.ets)
        self.Rhatee = np.mean(np.power(self.ets, 2))
        n = len(self.ets)
        self.tau = 1.96 * math.sqrt(self.Rhatee / n)
        self.iszeromean = True
        if abs(self.mhate) > self.tau:
            self.iszeromean = False
        self.zmw = self.autocorr1(self.ets)

    def plot(self):
        self.finalize()
        lw = 1
        plt.subplot(3, 2, 1), plt.plot(self.tts, self.log[:, 1], linewidth=lw), plt.ylabel('xhat')
        plt.subplot(3, 2, 2), plt.plot(self.tts, self.log[:, 3], linewidth=lw), plt.ylabel('err')
        plt.subplot(3, 2, 3), plt.plot(self.tts, self.log[:, 2], linewidth=lw), plt.ylabel('yhat')
        plt.subplot(3, 2, 4), plt.plot(self.tts, self.log[:, 4], linewidth=lw), plt.ylabel('innov')
        plt.subplot(3, 2, 5), plt.plot(self.tts, self.zmw, linewidth=lw), plt.ylabel('autocorr')
        plt.show()


class Dists():
    def __init__(self):
        self.log = []

    def update(self, t, tru, est):
        from pymc3.plots.kdeplot import fast_kde
        trupmf, trumin, trumax = fast_kde(tru)
        estpmf, estmin, estmax = fast_kde(est)
        truval = np.linspace(trumin, trumax, len(trupmf))
        estval = np.linspace(estmin, estmax, len(estpmf))
        trupmf = trupmf / sum(trupmf)
        estpmf = estpmf / sum(estpmf)
        self.log.append([t, truval, trupmf, estval, estpmf])

    def kld1(self, x1, x2):
        def kli(a, b): return np.sum(np.multiply(a, np.log(a)) - np.multiply(a, np.log(b)), axis=0)
        x1 = 1.0 * x1 / np.sum(x1, axis=0)
        x2 = 1.0 * x2 / np.sum(x2, axis=0)
        return (kli(x1, x2) + kli(x2, x1)) / 2

    def kld2(self, x1, x2):
        from scipy.stats import entropy
        return entropy(x1, x2)

    def plot(self, tlim = 100):
        import matplotlib.pyplot as plt
        est = []
        for rec in self.log: est.append([rec[0], rec[1][np.argmax(rec[2])], 0., rec[3][np.argmax(rec[4])], 0.])
        est = np.asarray(est)
        plt.figure()
        ax = plt.axes(projection='3d')
        ax.plot3D(est[:tlim,1], est[:tlim,0], est[:tlim,2], c='r', linewidth=1)
        ax.plot3D(est[:tlim,3], est[:tlim,0], est[:tlim,4], c='r', linewidth=1)
        for rec in self.log[:tlim]:
            ax.plot3D(rec[1], rec[0] * np.ones(len(rec[1])), rec[2], c='g', linewidth=1)
            ax.plot3D(rec[3], rec[0] * np.ones(len(rec[3])), rec[4], c='b', linewidth=1)
        plt.title('kl-divergence, %.1E' % (self.kld1(est[:tlim, 1], est[:tlim, 3])))
        ax.set_xlabel('x')
        ax.set_ylabel('t')
        ax.set_zlabel('p')
        plt.show()


class DistsEns():
    def __init__(self):
        self.log = []

    def update(self, distslog):
        for rec in distslog:
            truval, trupmf = rec[1], rec[2]
            maptru = truval[np.argmax(trupmf)]
            estval, estpmf = rec[3], rec[4]
            mapest = estval[np.argmax(estpmf)]
            self.log.append([maptru, mapest])

    def plot(self):
        log = np.asarray(self.log)
        from pymc3.plots.kdeplot import fast_kde
        import matplotlib.pyplot as plt
        maptrupmf, maptrumin, maptrumax = fast_kde(log[:, 0])
        mapestpmf, mapestmin, mapestmax = fast_kde(log[:, 1])
        maptruval = np.linspace(maptrumin, maptrumax, len(maptrupmf))
        mapestval = np.linspace(mapestmin, mapestmax, len(mapestpmf))
        maptrupmf = maptrupmf / sum(mapestpmf)
        mapestpmf = mapestpmf / sum(mapestpmf)
        plt.figure()
        plt.plot(maptruval, maptrupmf, 'g')
        plt.plot(mapestval, mapestpmf, 'b')
        plt.show()
        pass


import numpy as np
from decisions import Innovs
import statespacemodels

def ud_factorization(M):
    assert np.allclose(M, M.T)
    n = M.shape[0]
    M = np.triu(M)
    U = np.eye(n)
    d = np.zeros(n)
    for j in reversed(range(2, n + 1)):
        d[j - 1] = M[j - 1, j - 1]
        if d[j - 1] > 0:
            alpha = 1.0 / d[j - 1]
        else:
            alpha = 0.0
        for k in range(1, j):
            beta = M[k - 1, j - 1]
            U[k - 1, j - 1] = alpha * beta
            M[0:k, k - 1] = M[0:k, k - 1] - beta * U[0:k, j - 1]
    d[0] = M[0, 0]
    return U, np.diag(d)


def temporal_update(xin, Uin, Din, m):  # thornton
    Phi, Gin, Q = m.A(xin), m.G, m.Q
    x, U, D = Phi @ xin, Uin, Din
    n, r = 3, 3
    G = Gin
    U = np.eye(3)
    PhiU = Phi @ Uin
    for i in reversed(range(3)):
        sigma = 0
        for j in range(n):
            sigma = sigma + PhiU[i, j] ** 2 * Din[j, j]
            if (j <= r - 1):
                sigma = sigma + G[i, j] ** 2 + Q[j, j]
        D[i, i] = sigma
        ilim = i - 1
        if not ilim < 0:
            for j in range(ilim):
                sigma = 0
                for k in range(n):
                    sigma = sigma + PhiU[i, k] * Din[k, k] * PhiU[j, k]
                for k in range(r):
                    sigma = sigma + G[i, k] * Q[k, k] * G[j, k]
                U[j, i] = sigma / D[i, i]
                for k in range(n):
                    PhiU[j, k] = PhiU[j, k] - U[j, i] * PhiU[i, k]
                for k in range(r):
                    G[j, k] = G[j, k] - U[j, i] * G[i, k]
    return x, U, D


def observational_update(xin, Uin, Din, obs, m):  # bierman
    R, H, yhat = m.Rvv, m.C(xin), m.c(xin)
    x, U, D = xin, Uin, Din
    a = U.T @ H.T
    b = D @ a
    dz = obs - yhat  # z - H @ xin
    alpha = R
    gamma = 1 / alpha
    for j in range(3):
        beta = alpha
        alpha = alpha + a[j] * b[j]
        lamda = -a[j] * gamma
        gamma = 1 / alpha
        D[j, j] = beta * gamma * D[j, j]
        jlim = j - 1
        if not jlim < 0:
            for i in range(jlim):
                beta = U[i, j]
                U[i, j] = beta + b[i] * lamda
                b[i] = b[i] + b[j] * beta
    dzs = gamma * dz
    x = x + dzs * b
    return x, U, D, yhat


class Classical():
    '''classical kalman filter.'''
    def __init__(self, mode, plot=True, signal=None):
        self.innov = Innovs()
        if mode == 'rccircuit':
            if signal == None: signal = 300.
            m = statespacemodels.Rccircuit(signal=signal)
            self.rccircuit(m)
        elif mode == 'kf2':
            m = statespacemodels.Jazwinski1()
            self.kf2(m)
        elif mode == 'ekf1':
            m = statespacemodels.Jazwinski1()
            self.ekf1(m)
        elif mode == 'ekf2':
            m = statespacemodels.Jazwinski2()
            self.ekf2(m)
        if plot: self.innov.plot()

    def rccircuit(self, m):
        xhat = 2.5
        Ptil = 50e-4
        for step in m.steps():
            xhat = .97 * xhat + 100 * m.u
            Ptil = .94 * Ptil + m.Rww
            Ree = 4 * Ptil + 4
            K = 2 * Ptil / Ree
            yhat = 2 * xhat
            xhat = xhat + K * (step[2] - yhat)
            Ptil = Ptil / (Ptil + 1)
            self.innov.update2(step[0], xhat, yhat, step[1] - xhat, step[2] - yhat, Ree, Ptil)

    def kf2(self, m):
        xhat = 2.2
        Ptil = .01
        for step in m.steps():
            xref = 2. + .067 * step[0]
            xhat = m.a(xref, 0) + m.A(xref) * (xhat - xref)
            Ptil = m.A(xref) * Ptil * m.A(xref)
            Ree = m.C(xref) * Ptil * m.C(xref) + m.Rvv
            K = Ptil * m.C(xref) / Ree
            yhat = m.c(xref, 0) + m.C(xref) * (xhat - xref)
            xhat = xhat + K * (step[2] - yhat)
            Ptil = (1 - K * m.C(xref)) * Ptil
            self.innov.update(step[0], xhat, yhat, step[1] - xhat, step[2] - yhat)

    def ekf1(self, m):
        xhat = 2.2
        Ptil = .01
        for step in m.steps():
            xhat = m.a(xhat)
            Ptil = m.A(xhat) * Ptil * m.A(xhat)
            Ree = m.C(xhat) * Ptil * m.C(xhat) + m.Rvv
            K = Ptil * m.C(xhat) / Ree
            yhat = m.c(xhat)
            xhat = xhat + K * (step[2] - yhat)
            Ptil = (1 - K * m.C(xhat)) * Ptil
            self.innov.update(step[0], xhat, yhat, step[1] - xhat, step[2] - yhat)

    def ekf2(self, m):
        xhat = np.array([2, .055, .044])
        U, D = ud_factorization(1. * np.eye(3))
        for step in m.steps():
            xhat, U, D = temporal_update(xin=m.a(xhat), Uin=U, Din=D, m=m)
            xhat, U, D, yhat = observational_update(xin=xhat, Uin=U, Din=D, obs=step[2], m=m)
            self.innov.update(step[0], xhat[0], yhat, step[1][0] - xhat[0], step[2] - yhat)

def main():
    Classical('rccircuit', plot=True)
    # Classical('kf2')
    # Classical('ekf1')
    # Classical('ekf2')

if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt
import numpy as np
import math

def main():
    e1 = Exp1()
    e1.filtering()
    e1.state_smoothing()
    e1.simulation()
    e1.plot()
    plt.show()

# inverse-gamma distributions of the disturbance variances sigsqrepsilon and sigsqreta
class Params():
    def __init__(self):
        self.g1 = [3, .02]
        self.g2 = [3, .02]

    def path(self, epsilons, etas):
        self.theta1 = [1e-2, 1e-5]
        self.theta2 = [1e-2, 1e-5]
        N = len(epsilons)
        self.theta1[0] = self.theta1[0] + N / 2
        self.theta2[0] = self.theta2[0] + N / 2
        for j in range(1, N):
            self.theta1[1] += .5 * epsilons[j]**2
            self.theta2[1] += .5 * etas[j]**2

class Exp1():
    def __init__(self):
        self.ts = [float(rec[0]) for rec in nile()]
        self.y = [float(rec[1]) for rec in nile()]
        self.prep()
        self.sigsqrepsilon = 15099.
        self.sigsqreta = 1469.1
        self.a[0] = 1120.
        self.P[0] = 1e4

    def filtering(self):
        for t in range(len(self.y)):
            self.v[t] = self.y[t] - self.a[t]
            self.F[t] = self.P[t] + self.sigsqrepsilon
            self.K[t] = self.P[t] / self.F[t]
            if t < len(self.y) - 1:
                self.a[t+1] = self.a[t] + self.K[t] * self.v[t]
                self.P[t+1] = self.P[t] * (1 - self.K[t]) + self.sigsqreta

    def state_smoothing(self):
        for t in range(len(self.y) - 1, -1, -1):
            self.r[t-1] = (1 / self.F[t]) * self.v[t] + (1 - self.K[t]) * self.r[t]
            self.alphahat[t] = self.a[t] + self.P[t] * self.r[t-1]
            self.N[t-1] = (1 / self.F[t]) + (1 - self.K[t])**2 * self.N[t]
            self.V[t] = self.P[t] - self.P[t]**2 * self.N[t-1]

    def disturbance_smoothing(self, y):
        v = np.zeros(len(y))
        r = np.zeros(len(y))
        epsilonhat = np.zeros(len(y))
        for t in range(len(v)): v[t] = y[t] - self.a[t]
        for t in range(len(r) - 1, -1, -1): r[t-1] = (1 / self.F[t]) * v[t] + (1 - self.K[t]) * r[t]
        for t in range(len(epsilonhat)):
            u = (1 / self.F[t]) * v[t] - self.K[t] * r[t]
            epsilonhat[t] = self.sigsqrepsilon * u
        return epsilonhat

    def simulation(self):
        epsilonhat = self.disturbance_smoothing(self.y)
        etaplus = math.sqrt(self.sigsqreta) * np.random.normal(0., 1., 100)
        epsilonplus = math.sqrt(self.sigsqrepsilon) * np.random.normal(0., 1., 100)
        alphaplus = np.zeros(len(self.y))
        yplus = np.zeros(len(self.y))
        alphaplus[0] = self.y[0]
        for t in range(len(self.y)):
            yplus[t] = alphaplus[t] + epsilonplus[t]
            if t < len(self.y) - 1:
                alphaplus[t + 1] = alphaplus[t] + etaplus[t]
        epsilonplushat = self.disturbance_smoothing(yplus)
        for t in range(len(self.y)):
            self.epsilontil[t] = epsilonplus[t] + (epsilonhat[t] - epsilonplushat[t])
            self.alphatil[t] = self.y[t] - self.epsilontil[t]
            if t < len(self.y) - 1:
                self.etatil[t] = self.alphatil[t+1] - self.alphatil[t]

    def prep(self):
        self.a = np.zeros(len(self.y))
        self.P = np.zeros(len(self.y))
        self.v = np.zeros(len(self.y))
        self.F = np.zeros(len(self.y))
        self.K = np.zeros(len(self.y))
        self.r = np.zeros(len(self.y))
        self.alphahat = np.zeros(len(self.y))
        self.N = np.zeros(len(self.y))
        self.V = np.zeros(len(self.y))
        self.epsilontil = np.zeros(len(self.y))
        self.alphatil = np.zeros(len(self.y))
        self.etatil = np.zeros(len(self.y))

    def plot(self):
        plt.figure()
        plt.plot(self.ts, self.a, c='b', linewidth=1, alpha=.25)
        plt.plot(self.ts, self.a + 2. * np.sqrt(np.asarray(self.P)), c='b', linewidth=1, alpha=.1)
        plt.plot(self.ts, self.a - 2. * np.sqrt(np.asarray(self.P)), c='b', linewidth=1, alpha=.1)
        plt.plot(self.ts, self.alphahat, c='g', linewidth=1)
        plt.plot(self.ts, self.alphahat + 2. * np.sqrt(np.asarray(self.V)), c='g', linewidth=1, alpha=.5)
        plt.plot(self.ts, self.alphahat - 2. * np.sqrt(np.asarray(self.V)), c='g', linewidth=1, alpha=.5)
        plt.scatter(self.ts, self.alphatil, c='g', s=5) # sample from state dist conditional on obs
        plt.scatter(self.ts, self.y, c='b', s=5, alpha=.25)

def nile(): return [[1871, 1120], [1872, 1160], [1873, 963], [1874, 1210], [1875, 1160], [1876, 1160], [1877, 813],
              [1878, 1230], [1879, 1370], [1880, 1140], [1881, 995], [1882, 935], [1883, 1110], [1884, 994],
              [1885, 1020], [1886, 960], [1887, 1180], [1888, 799], [1889, 958], [1890, 1140], [1891, 1100],
              [1892, 1210], [1893, 1150], [1894, 1250], [1895, 1260], [1896, 1220], [1897, 1030], [1898, 1100],
              [1899, 774], [1900, 840], [1901, 874], [1902, 694], [1903, 940], [1904, 833], [1905, 701],
              [1906, 916], [1907, 692], [1908, 1020], [1909, 1050], [1910, 969], [1911, 831], [1912, 726],
              [1913, 456], [1914, 824], [1915, 702], [1916, 1120], [1917, 1100], [1918, 832], [1919, 764],
              [1920, 821], [1921, 768], [1922, 845], [1923, 864], [1924, 862], [1925, 698], [1926, 845],
              [1927, 744], [1928, 796], [1929, 1040], [1930, 759], [1931, 781], [1932, 865], [1933, 845],
              [1934, 944], [1935, 984], [1936, 897], [1937, 822], [1938, 1010], [1939, 771], [1940, 676],
              [1941, 649], [1942, 846], [1943, 812], [1944, 742], [1945, 801], [1946, 1040], [1947, 860],
              [1948, 874], [1949, 848], [1950, 890], [1951, 744], [1952, 749], [1953, 838], [1954, 1050],
              [1955, 918], [1956, 986], [1957, 797], [1958, 923], [1959, 975], [1960, 815], [1961, 1020],
              [1962, 906], [1963, 901], [1964, 1170], [1965, 912], [1966, 746], [1967, 919], [1968, 718],
              [1969, 714], [1970, 740]]

if __name__ == "__main__":
    main()



import math
import numpy as np
from scipy.linalg.blas import drot, drotg
from decisions import Innovs
import statespacemodels

def cholupdate(R, z):
    n = z.shape[0]
    for k in range(n):
        c, s = drotg(R[k, k], z[k])
        drot(R[k, :], z, c, s, overwrite_x=True, overwrite_y=True)
    return R

def choldowndate(R, z):
    n = R.shape[0]
    for k in range(n):
        if (R[k, k] - z[k]) * (R[k, k] + z[k]) < 0: return R
        rbar = np.sqrt((R[k, k] - z[k]) * (R[k, k] + z[k]))
        for j in range(k + 1, n):
            R[k, j] = 1. / rbar * (R[k, k] * R[k, j] - z[k] * z[j])
            z[j] = 1. / R[k, k] * (rbar * z[j] - z[k] * R[k, j])
        R[k, k] = rbar
    return R

def temporal_update(xhat, S, m):
    X = m.va(m.X(xhat, S))
    xhat = m.Wm @ X.T
    for i in range(7): m.Xtil[:, i] = X[:, i] - xhat
    q, r = np.linalg.qr(np.concatenate([math.sqrt(m.Wc[1]) * m.Xtil[:, 1:], m.Sw], 1))
    S = cholupdate(r.T[0:3, 0:3], m.Wc[0] * m.Xtil[:, 0])
    return xhat, S, X

def observational_update(xhat, S, X, obs, m):
    Y = m.vc(m.Xhat(X))
    yhat = m.Wm @ Y.T
    for i in range(7): m.Ytil[0, i] = Y[i] - yhat
    q, r = np.linalg.qr(np.concatenate([math.sqrt(m.Wc[1]) * m.Ytil[:, 1:], m.Sv], 1))
    Sy = cholupdate(r.T[0:1, 0:1], m.Wc[0] * m.Ytil[:, 0])
    for i in range(7): m.Pxy[:, 0] = m.Pxy[:, 0] + m.Wc[i] * m.Xtil[:, i] * m.Ytil.T[i, :]
    if Sy[0, 0] < math.sqrt(10) or Sy[0, 0] > math.sqrt(1000): Sy[0, 0] = math.sqrt(1000)
    K = m.Pxy / Sy[0, 0] ** 2
    U = K * Sy
    xhat = xhat + K[:, 0] * (obs - yhat)
    S = choldowndate(S, U)
    return xhat, S, yhat

class Modern():
    def __init__(self, mode, plot=True):
        self.innov = Innovs()
        if mode == 'spkf1':
            m = statespacemodels.Jazwinski1()
            self.spkf1(m)
        elif mode == 'spkf2':
            m = statespacemodels.Jazwinski2()
            self.spkf2(m)
        if plot: self.innov.plot()

    def spkf1(self, m):
        xhat = 2.2
        Ptil = .01
        for step in m.steps():
            X = m.va(m.X(xhat, Ptil))
            Ptil = m.W @ np.power(X - m.W @ X, 2) + m.Rww
            Y = m.vc(m.Xhat(X, m.Rww))
            Rksiksi = m.W @ np.power(Y - m.W @ Y, 2) + m.Rvv
            RXtilksi = m.W @ np.multiply(X - m.W @ X, Y - m.W @ Y)
            K = RXtilksi / Rksiksi
            yhat = m.W @ Y
            xhat = m.W @ X + K * (step[2] - m.W @ Y)
            Ptil = Ptil - K * Rksiksi * K
            self.innov.update(step[0], xhat, yhat, step[1] - xhat, step[2] - yhat)

    def spkf2(self, m):
        xhat = np.array([2.0, .055, .044])
        S = np.linalg.cholesky(.1 * np.eye(3))
        for step in m.steps():
            xhat, S, X = temporal_update(xhat=xhat, S=S, m=m)
            xhat, S, yhat = observational_update(xhat=xhat, S=S, X=X, obs=step[2], m=m)
            self.innov.update(step[0], xhat[0], yhat, step[1][0] - xhat[0], step[2] - yhat)

if __name__ == "__main__":
    Modern('spkf1')
    Modern('spkf2')



import setuptools, os

__version__ = None
if os.environ.get('CI_COMMIT_TAG'):
    __version__ = os.environ['CI_COMMIT_TAG']
elif os.environ.get('CI_JOB_ID'):
    __version__ = os.environ['CI_JOB_ID']
if __version__:
    with open('__version__', 'wt') as fout:
        fout.write(__version__)
if os.path.exists('__version__'):
    with open('__version__', 'rt') as fin:
        for line in fin:
            __version__ = line.strip()

with open('requirements.txt') as fin:
    required = fin.read().splitlines()
with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='statespace',
    version=__version__,
    author='noah smith',
    author_email='noahhsmith@gmail.com',
    description='state-space distributions and decisions',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/noahhsmith/statespace',
    packages=setuptools.find_packages(),
    entry_points={'console_scripts': ['statespace = statespace.__main__:main']},
    install_requires=required,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
